<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Controller;

use App\Form\AgentConfigurationForm;
use App\itnovum\openITCOCKPIT\Monitoring\Naemon\ExternalCommands;
use App\Model\Entity\Changelog;
use App\Model\Entity\Host;
use App\Model\Table\AgentconfigsTable;
use App\Model\Table\ChangelogsTable;
use App\Model\Table\HostsTable;
use App\Model\Table\HosttemplatesTable;
use App\Model\Table\PushAgentsTable;
use App\Model\Table\ServicesTable;
use App\Model\Table\ServicetemplatesTable;
use Cake\Datasource\Exception\RecordNotFoundException;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\I18n\FrozenTime;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Agent\AgentConfiguration;
use itnovum\openITCOCKPIT\Agent\AgentHttpClient;
use itnovum\openITCOCKPIT\Agent\AgentResponseToServices;
use itnovum\openITCOCKPIT\Core\AngularJS\Api;
use itnovum\openITCOCKPIT\Core\Comparison\ServiceComparisonForSave;
use itnovum\openITCOCKPIT\Core\HostConditions;
use itnovum\openITCOCKPIT\Core\System\Gearman;
use itnovum\openITCOCKPIT\Core\Views\ContainerPermissions;
use itnovum\openITCOCKPIT\Filter\HostFilter;

/**
 *
 * Class AgentconnectorController
 *
 *
 *
 * -------------------------> !!! READ THIS !!! <-------------------------
 * This file has to keep in sync with: https://github.com/it-novum/openITCOCKPIT/blob/development/src/Controller/AgentconnectorController.php
 * IF YOU MAKE A CHANGE IN THIS FILE MAKE SURE TO CHANGE BOTH
 * The files are slightly different on the Master and on the Satellite system - Please do not just copy & paste them over.
 *
 * This version only contains the required methods for openITCOCKPIT Monitoring Agents running in Push Mode
 */
class AgentconnectorController extends AppController {

    /************************************
     *    AGENT API METHODS FOR PUSH    *
     ************************************/

    /**
     * Register new PUSH Agents
     *
     * How it works:
     * Register new Agents:
     * The Agent send it's UUID and an empty Password to the openITCOCKPIT Server. If no password was generated for the given UUID
     * openITCOCKPIT will generate a new Password and respond this password to the Agent. Respond with 201
     *
     * If an Agent sends an password which is not found in the database, openITCOCKPIT will respond with an 403 Forbidden
     */
    public function register_agent() {
        $agentUuid = $this->request->getData('agentuuid', null);
        $agentPassword = $this->request->getData('password', null);

        if (!$this->request->is('post') || !$this->isJsonRequest()) {
            throw new MethodNotAllowedException();
        }

        if ($agentUuid === null || $agentPassword === null) {
            $this->response = $this->response->withStatus(400);
            $this->set('error', 'Field uuid or password is missing in POST data');
            $this->viewBuilder()->setOption('serialize', ['error']);
            return;
        }

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        if ($agentPassword === '' && $PushAgentsTable->existsByUuid($agentUuid)) {
            // It this UUID already registered?
            $this->response = $this->response->withStatus(403);
            $this->set('error', 'The given UUID is already registed with a password!');
            $this->viewBuilder()->setOption('serialize', ['error']);
            return;
        }

        if ($agentPassword === '' && !$PushAgentsTable->existsByUuid($agentUuid)) {
            // New or unknown agent - Create a new password for this Agent and add it to our database
            $bytes = openssl_random_pseudo_bytes(64, $cstrong);
            $password = bin2hex($bytes);

            $remoteAddress = null;
            if (!empty($_SERVER['REMOTE_ADDR'])) {
                $remoteAddress = $_SERVER['REMOTE_ADDR'];
            }
            if (!empty($_SERVER['HTTP_CLIENT_IP'])) {
                $remoteAddress = $_SERVER['HTTP_CLIENT_IP'];
            }
            $HTTP_X_FORWARDED_FOR = null;
            if (!empty($_SERVER['HTTP_X_FORWARDED_FOR'])) {
                $HTTP_X_FORWARDED_FOR = $_SERVER['HTTP_X_FORWARDED_FOR'];
            }


            $entity = $PushAgentsTable->newEntity([
                'uuid'                 => $agentUuid,
                'host_uuid'            => null,
                'password'             => $password,
                'hostname'             => $this->request->getData('hostname', null),
                'ipaddress'            => $this->request->getData('ipaddress', null),
                'remote_address'       => $remoteAddress,
                'http_x_forwarded_for' => $HTTP_X_FORWARDED_FOR,
                'last_update'          => new FrozenTime(),
                'checkresults'         => null
            ]);

            $PushAgentsTable->save($entity);
            if ($entity->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $entity->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            //Send new Password to Agent
            $this->response = $this->response->withStatus(201);
            $this->set('agentuuid', $agentUuid);
            $this->set('password', $password);
            $this->viewBuilder()->setOption('serialize', ['agentuuid', 'password']);
            return;
        }

        // Password and Agent UUID given - check if this exists in the database
        if ($PushAgentsTable->existsByUuidAndPassword($agentUuid, $agentPassword)) {
            $this->response = $this->response->withStatus(200);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['success']);
            return;
        }

        $this->response = $this->response->withStatus(404);
        $this->set('error', 'No Agent found for given UUID and password');
        $this->viewBuilder()->setOption('serialize', ['error']);
        return;
    }

    /**
     * Receiver function called by Agents running in PUSH mode
     */
    public function submit_checkdata() {
        if (!$this->isJsonRequest() || !$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        $agentUuid = $this->request->getData('agentuuid', '');
        $agentPassword = $this->request->getData('password', '');
        $checkdata = $this->request->getData('checkdata', []);

        /** @var PushAgentsTable $PushAgentsTable */
        $PushAgentsTable = TableRegistry::getTableLocator()->get('PushAgents');

        $receivedChecks = 0;
        try {
            $pushAgent = $PushAgentsTable->getConfigWithHostForSubmitCheckdata(
                $agentUuid,
                $agentPassword
            );

            $hostUuid = $pushAgent->get('host_uuid');

            // Agent is known and authorized
            require_once "/opt/openitc/receiver/vendor/autoload.php";
            $CheckConfig = new \itnovum\openITCOCKPIT\Checks\Receiver\CheckConfig('/opt/openitc/receiver/etc/production.json');
            try {
                $config = $CheckConfig->getConfigByHostName($hostUuid);

                $results = [
                    'Messages' => []
                ];

                foreach ($config['checks'] as $pluginConfig) {
                    $pluginName = $pluginConfig['plugin'];

                    $pluginClassName = sprintf('itnovum\openITCOCKPIT\Checks\Receiver\Plugins\%s', $pluginName);
                    if (!class_exists($pluginClassName)) {
                        //Unknown Plugin
                        continue;
                    }

                    /** @var  $Plugin \itnovum\openITCOCKPIT\Checks\Receiver\Plugins\PluginInterface */
                    $Plugin = new $pluginClassName($pluginConfig, $checkdata);

                    $pluginOutput = $Plugin->getOutput();
                    if (strlen($Plugin->getPerfdataSerialized()) > 0) {
                        $pluginOutput .= '|' . $Plugin->getPerfdataSerialized();
                    }

                    // Create bulk message for Statusengine Broker
                    $results['messages'][] = [
                        'Command' => 'check_result',
                        'Data'    => [
                            'host_name'           => $hostUuid,
                            'service_description' => $pluginConfig['uuid'],
                            'output'              => $pluginOutput,
                            'long_output'         => ($Plugin->getLongOutput() === '' ? null : $Plugin->getLongOutput()),
                            'check_type'          => 1, //https://github.com/naemon/naemon-core/blob/cec6e10cbee9478de04b4cf5af29e83d47b5cfd9/src/naemon/common.h#L330-L334
                            'return_code'         => $Plugin->getStatuscode(),
                            'start_time'          => time(),
                            'end_time'            => time(),
                            'early_timeout'       => 0,
                            'latency'             => 0,
                            'exited_ok'           => 1
                        ]
                    ];

                    $receivedChecks++;
                }

                if (!empty($results['messages'])) {
                    $ExternalCommands = new ExternalCommands();
                    $ExternalCommands->toStatusnginCmdBackground($results);
                }

            } catch (\RuntimeException $e) {
                // Host was not exported yet to Monitoring Engine and check receiver - just save the check result to the database
                // and ignore the error
            }

            $pushAgent->set('last_update', new FrozenTime());
            $pushAgent->set('checkresults', json_encode($checkdata));
            $PushAgentsTable->save($pushAgent);

            $this->set('success', true);
            $this->set('received_checks', $receivedChecks);
            $this->viewBuilder()->setOption('serialize', ['success', 'received_checks']);
            return;
        } catch (RecordNotFoundException $e) {
            //Agent has no host_uuid assignment yet. Due to we are on the satellite, we store the check results anyway
            //so we can answer to possible requests from the master system.

            try {
                $pushAgent = $PushAgentsTable->getRecordByAgentUuidAndPassword($agentUuid, $agentPassword);
                $pushAgent->set('last_update', new FrozenTime());
                $pushAgent->set('checkresults', json_encode($checkdata));
                $PushAgentsTable->save($pushAgent);
            } catch (RecordNotFoundException $e) {
                // Unknown agent - we dont save anything
            }
        }

        $this->response = $this->response->withStatus(400);
        $this->set('error', 'Invalid credentials or host not found');
        $this->viewBuilder()->setOption('serialize', ['error']);
    }
}
