<?php
/**
 * CakePHP(tm) : Rapid Development Framework (http://cakephp.org)
 * Copyright (c) Cake Software Foundation, Inc. (http://cakefoundation.org)
 *
 * Licensed under The MIT License
 * For full copyright and license information, please see the LICENSE.txt
 * Redistributions of files must retain the above copyright notice.
 *
 * @copyright Copyright (c) Cake Software Foundation, Inc. (http://cakefoundation.org)
 * @link      http://cakephp.org CakePHP(tm) Project
 * @since     0.2.9
 * @license   http://www.opensource.org/licenses/mit-license.php MIT License
 */

declare(strict_types=1);

namespace App\Controller;

use App\itnovum\openITCOCKPIT\Core\Permissions\MyRightsFactory;
use Authentication\Controller\Component\AuthenticationComponent;
use Authentication\IdentityInterface;
use Cake\Cache\Cache;
use Cake\Controller\Controller;
use Cake\Event\Event;
use Cake\Event\EventInterface;
use Cake\Http\Response;
use Exception;

/**
 * Class AppController
 * @package App\Controller
 * @property AuthenticationComponent $Authentication
 */
class AppController extends Controller {

    /**
     * @var array
     */
    protected $PERMISSIONS = [];


    /**
     * Initialization hook method.
     *
     * Use this method to add common initialization code like loading components.
     *
     * e.g. `$this->loadComponent('Security');`
     *
     * @return void
     * @throws Exception
     */
    public function initialize(): void {
        parent::initialize();

        $this->loadComponent('Flash');
        $this->loadComponent('RequestHandler', [
            'enableBeforeRedirect' => false
        ]);

        // Docs: https://book.cakephp.org/authentication/1/en/index.html
        $this->loadComponent('Authentication.Authentication', [
            'logoutRedirect' => '/users/login'  // Default is false
        ]);
    }

    public function beforeFilter(EventInterface $event) {
        parent::beforeFilter($event);


        $user = $this->Authentication->getIdentity();

        if ($user !== null) {
            $userId = $user->get('id');

            //User is logged in
            $cacheKey = 'userPermissions_' . $userId;

            if (Cache::read($cacheKey, 'permissions') === null) {
                $permissions = $this->getUserPermissions($user);

                Cache::write($cacheKey, $permissions, 'permissions');
            }

            $permissions = Cache::read($cacheKey, 'permissions');

            $this->PERMISSIONS = $permissions['PERMISSIONS'];
        }
    }

    /**
     * @param IdentityInterface $identity
     * @return array
     */
    private function getUserPermissions(IdentityInterface $identity) {
        $userId = $identity->get('id');
        $usergroupId = $identity->get('usergroup_id');
        $userPermissions = MyRightsFactory::getUserPermissions($userId, $usergroupId);
        return $userPermissions;
    }

    /**
     * Before render callback.
     *
     * @param Event $event The beforeRender event.
     * @return Response|null|void
     */
    public function beforeRender(EventInterface $event) {
        if (!$this->isApiRequest()) {
            //Set Permissions for ACL Helper
            $this->set('ACLPERMISSIONS', $this->PERMISSIONS);
        }
        if ($this->isJsonRequest()) {
            $this->set('_csrfToken', $this->request->getAttribute('csrfToken'));
            $serialize = $this->viewBuilder()->getOption('serialize');
            if ($serialize === null) {
                $serialize = [];
            }
            $serialize[] = '_csrfToken';
            //Add Paginator info to json response
            $paging = $this->viewBuilder()->getVar('paging');
            if ($paging !== null) {
                $serialize[] = 'paging';
            }

            //Add Scroll Paginator info to json response
            $scroll = $this->viewBuilder()->getVar('scroll');
            if ($scroll !== null) {
                $serialize[] = 'scroll';
            }
            $this->viewBuilder()->setOption('serialize', $serialize);
        }
    }

    /**
     * @return bool
     */
    protected function isApiRequest(): bool {
        if ($this->isJsonRequest() || $this->isXmlRequest()) {
            return true;
        }

        return false;
    }

    /**
     * @return bool
     */
    protected function isJsonRequest(): bool {
        return $this->request->getParam('_ext') === 'json';
    }

    /**
     * @return bool
     */
    protected function isXmlRequest(): bool {
        return $this->request->getParam('_ext') === 'xml';
    }

    /**
     * @return IdentityInterface|null
     */
    protected function getUser() {
        return $this->Authentication->getIdentity();
    }

    protected function isAngularJsRequest() {
        if ($this->isApiRequest()) {
            return $this->request->getQuery('angular') !== null;
        }
        return false;
    }

    protected function isScrollRequest() {
        if ($this->isApiRequest()) {
            if ($this->request->getQuery('scroll', 'false') !== 'false') {
                return true;
            }
        }

        return false;
    }

    /**
     * @param string|null $action
     * @param string|null $controller
     * @param string|null $plugin
     * @return bool
     */
    protected function hasPermission($action = null, $controller = null, $plugin = null) {
        $controller = strtolower($controller);
        $action = strtolower($action);
        if ($plugin !== null) {
            $plugin = strtolower($plugin);
        }

        //return false;
        if ($plugin === null) {
            $plugin = Inflector::classify($this->getRequest()->getParam('plugin', ''));
        }

        if ($plugin === null || $plugin === '') {
            return isset($this->PERMISSIONS[$controller][$action]);
        }

        return isset($this->PERMISSIONS[$plugin][$controller][$action]);
    }
}
