<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace App\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Entity\DowntimeHost;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\DowntimeHostConditions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;

/**
 * DowntimeHost Model
 *
 * @method \Model\Entity\DowntimeHost newEmptyEntity()
 * @method \Model\Entity\DowntimeHost newEntity(array $data, array $options = [])
 * @method \Model\Entity\DowntimeHost[] newEntities(array $data, array $options = [])
 * @method \Model\Entity\DowntimeHost get($primaryKey, $options = [])
 * @method \Model\Entity\DowntimeHost findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \Model\Entity\DowntimeHost patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \Model\Entity\DowntimeHost[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \Model\Entity\DowntimeHost|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \Model\Entity\DowntimeHost saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \Model\Entity\DowntimeHost[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeHost[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeHost[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \Model\Entity\DowntimeHost[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 */
class DowntimeHostsTable extends Table {

    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('statusengine_host_downtimehistory');
        $this->setDisplayField('hostname');
        $this->setPrimaryKey(['hostname', 'node_name', 'scheduled_start_time', 'internal_downtime_id']);
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        //Readonly table
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        //Readonly table
        return $rules;
    }

    /**
     * @param DowntimeHostConditions $DowntimeHostConditions
     * @param PaginateOMat|null $PaginateOMat
     * @return array|void
     */
    public function getDowntimes(DowntimeHostConditions $DowntimeHostConditions, $PaginateOMat = null) {
        $query = $this->find()
            ->select([
                'DowntimeHosts.hostname',
                'DowntimeHosts.author_name',
                'DowntimeHosts.comment_data',
                'DowntimeHosts.entry_time',
                'DowntimeHosts.scheduled_start_time',
                'DowntimeHosts.scheduled_end_time',
                'DowntimeHosts.duration',
                'DowntimeHosts.was_started',
                'DowntimeHosts.internal_downtime_id',
                'DowntimeHosts.was_cancelled'
            ])
            ->where([
                'DowntimeHosts.scheduled_start_time >' => $DowntimeHostConditions->getFrom(),
                'DowntimeHosts.scheduled_start_time <' => $DowntimeHostConditions->getTo(),
            ])
            ->order($DowntimeHostConditions->getOrder());
        if ($DowntimeHostConditions->hideExpired()) {
            $query->andWhere([
                'DowntimeHosts.scheduled_end_time >' => time()
            ]);
        }

        if ($DowntimeHostConditions->hasConditions()) {
            $query->andWhere($DowntimeHostConditions->getConditions());
        }

        if ($DowntimeHostConditions->isRunning()) {
            $query->andWhere([
                'DowntimeHosts.scheduled_end_time >' => time(),
                'DowntimeHosts.was_started'          => 1,
                'DowntimeHosts.was_cancelled'        => 0
            ]);
        }

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }


    /**
     * @param int $internalDowntimeId
     * @return array
     */
    public function getHostUuidWithDowntimeByInternalDowntimeId($internalDowntimeId) {
        $result = $this->find()
            ->select([
                'DowntimeHosts.hostname',
                'DowntimeHosts.author_name',
                'DowntimeHosts.comment_data',
                'DowntimeHosts.entry_time',
                'DowntimeHosts.scheduled_start_time',
                'DowntimeHosts.scheduled_end_time',
                'DowntimeHosts.duration',
                'DowntimeHosts.was_started',
                'DowntimeHosts.internal_downtime_id',
                'DowntimeHosts.was_cancelled'
            ])
            ->where([
                'DowntimeHosts.internal_downtime_id' => $internalDowntimeId
            ])
            ->disableHydration()
            ->first();

        if ($result === null) {
            return [];
        }

        return $result;
    }

    /**
     * @param null $uuid
     * @param bool $isRunning
     * @return DowntimeHost|null
     */
    public function byHostUuid($uuid = null, $isRunning = false) {
        if (empty($uuid)) {
            return null;
        }

        $query = $this->find();
        $query->select([
            'DowntimeHosts.author_name',
            'DowntimeHosts.comment_data',
            'DowntimeHosts.entry_time',
            'DowntimeHosts.scheduled_start_time',
            'DowntimeHosts.scheduled_end_time',
            'DowntimeHosts.duration',
            'DowntimeHosts.was_started',
            'DowntimeHosts.internal_downtime_id',
            'DowntimeHosts.was_cancelled',
        ])
            ->order([
                'DowntimeHosts.entry_time' => 'DESC'
            ])
            ->where([
                'DowntimeHosts.hostname' => $uuid
            ]);

        if ($isRunning) {
            $query->andWhere([
                'DowntimeHosts.scheduled_end_time >' => time(),
                'DowntimeHosts.was_started'          => 1,
                'DowntimeHosts.was_cancelled'        => 0

            ]);
        }

        return $query->first();
    }
}
