<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.
declare(strict_types=1);

namespace App\Model\Table;

use App\Model\Entity\Servicegroup;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;

/**
 * Servicegroups Model
 *
 * @property \App\Model\Table\ServicegroupMembersTable&\Cake\ORM\Association\HasMany $ServicegroupMembers
 *
 * @method \App\Model\Entity\Servicegroup newEmptyEntity()
 * @method \App\Model\Entity\Servicegroup newEntity(array $data, array $options = [])
 * @method \App\Model\Entity\Servicegroup[] newEntities(array $data, array $options = [])
 * @method \App\Model\Entity\Servicegroup get($primaryKey, $options = [])
 * @method \App\Model\Entity\Servicegroup findOrCreate($search, ?callable $callback = null, $options = [])
 * @method \App\Model\Entity\Servicegroup patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method \App\Model\Entity\Servicegroup[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method \App\Model\Entity\Servicegroup|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \App\Model\Entity\Servicegroup saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method \App\Model\Entity\Servicegroup[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method \App\Model\Entity\Servicegroup[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method \App\Model\Entity\Servicegroup[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method \App\Model\Entity\Servicegroup[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 */
class ServicegroupsTable extends Table {
    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('servicegroups');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->hasMany('ServicegroupMembers', [
            'foreignKey' => 'servicegroup_id',
        ]);

    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('id')
            ->allowEmptyString('id', null, 'create');

        $validator
            ->scalar('uuid')
            ->maxLength('uuid', 37)
            ->requirePresence('uuid', 'create')
            ->notEmptyString('uuid')
            ->add('uuid', 'unique', ['rule' => 'validateUnique', 'provider' => 'table']);


        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->isUnique(['uuid']));

        return $rules;
    }

    /**
     * @param string $servicegroupUuid
     * @return Servicegroup
     */
    public function getServicessByServicegroupForMaps($servicegroupUuid) {
        $servicegroup = $this->find()
            ->select([
                'Servicegroups.id',
                'Servicegroups.uuid',
                'Servicegroups.description',
                'ServicegroupMapping.display_name'
            ])
            ->contain([
                'ServicegroupMembers' => function (Query $q) {
                    $q->select([
                        'ServicegroupMembers.id',
                        'ServicegroupMembers.servicegroup_id',
                        'ServicegroupMembers.service_uuid',
                        'ServiceMapping.display_name'
                    ])
                        ->join([
                            [
                                'table'      => 'mapping',
                                'alias'      => 'ServiceMapping',
                                'type'       => 'INNER',
                                'conditions' => [
                                    'ServicegroupMembers.service_uuid = ServiceMapping.uuid',
                                ]
                            ]
                        ]);
                    return $q;
                }
            ])
            ->join([
                [
                    'table'      => 'mapping',
                    'alias'      => 'ServicegroupMapping',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Servicegroups.uuid = ServicegroupMapping.uuid',
                    ],
                ],
            ])
            ->where([
                'Servicegroups.uuid' => $servicegroupUuid
            ])
            ->disableHydration()
            ->first();

        return $servicegroup ?? [];
    }

    /**
     * @param string $servicegroupUuid
     * @return Servicegroup
     */
    public function getServicessByServicegroupForMapsWithHost($servicegroupUuid) {
        $servicegroup = $this->find()
            ->select([
                'Servicegroups.id',
                'Servicegroups.uuid',
                'Servicegroups.description',
                'ServicegroupMapping.display_name'
            ])
            ->contain([
                'ServicegroupMembers' => function (Query $q) {
                    $q->select([
                        'ServicegroupMembers.id',
                        'ServicegroupMembers.servicegroup_id',
                        'ServicegroupMembers.service_uuid',
                        'ServiceMapping.display_name',
                        'Servicestatus.hostname',
                        'Servicestatus.service_description',
                        'HostMapping.display_name'
                    ])
                        ->join([
                            [
                                'table'      => 'mapping',
                                'alias'      => 'ServiceMapping',
                                'type'       => 'INNER',
                                'conditions' => [
                                    'ServicegroupMembers.service_uuid = ServiceMapping.uuid',
                                ]
                            ],
                            [
                                'table'      => 'statusengine_servicestatus',
                                'alias'      => 'Servicestatus',
                                'type'       => 'INNER',
                                'conditions' => [
                                    'ServicegroupMembers.service_uuid = Servicestatus.service_description',
                                ]
                            ],
                            [
                                'table'      => 'mapping',
                                'alias'      => 'HostMapping',
                                'type'       => 'INNER',
                                'conditions' => [
                                    'Servicestatus.hostname = HostMapping.uuid',
                                ]
                            ],
                        ]);
                    return $q;
                }
            ])
            ->join([
                [
                    'table'      => 'mapping',
                    'alias'      => 'ServicegroupMapping',
                    'type'       => 'INNER',
                    'conditions' => [
                        'Servicegroups.uuid = ServicegroupMapping.uuid',
                    ],
                ],
            ])
            ->where([
                'Servicegroups.uuid' => $servicegroupUuid
            ])
            ->disableHydration()
            ->firstOrFail();

        return $servicegroup ?? [];
    }
}
