<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace itnovum\openITCOCKPIT\Perfdata;


use App\Model\Table\PerfdataTable;
use App\Model\Table\ServicestatusTable;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\PerfdataBackend;
use itnovum\openITCOCKPIT\Graphite\GraphiteConfig;
use itnovum\openITCOCKPIT\Graphite\GraphiteLoader;
use itnovum\openITCOCKPIT\Graphite\GraphiteMetric;
use Statusengine\PerfdataParser;

/**
 * Class PerfdataLoader
 * @package itnovum\openITCOCKPIT\Perfdata
 */
class PerfdataLoader {

    const MAX_RESPONSE_GRAPH_POINTS = 1000;

    /**
     * @var PerfdataBackend
     */
    private $PerfdataBackend;

    /**
     * PerfdataLoader constructor.
     * @param PerfdataBackend $PerfdataBackend
     * @param null|string $gauge
     */
    public function __construct(PerfdataBackend $PerfdataBackend) {
        $this->PerfdataBackend = $PerfdataBackend;
    }


    /**
     * @param string $hostUuid
     * @param string $serviceUuid
     * @param int $start
     * @param int $end
     * @param bool $jsTimestamp
     * @param string $type
     * @param null $gauge
     * @param bool $scale
     * @param string|null $forcedUnit
     * @param bool $debug
     * @return array
     * @throws \GuzzleHttp\Exception\GuzzleException
     */
    public function getPerfdataByUuid($hostUuid, $serviceUuid, $start, $end, $jsTimestamp = false, $type = 'avg', $gauge = null, $scale = true, $forcedUnit = null, $debug = false) {
        if ($gauge === '') {
            $gauge = null;
        }

        $performance_data = [];

        /** @var ServicestatusTable $ServicestatusTable */
        $ServicestatusTable = TableRegistry::getTableLocator()->get('Servicestatus');
        $servicestatus = $ServicestatusTable->byUuid($serviceUuid);

        if ($this->PerfdataBackend->isWhisper()) {
            if (!empty($servicestatus)) {

                $GraphiteConfig = new GraphiteConfig();
                $GraphiteLoader = new GraphiteLoader($GraphiteConfig, $debug);

                if (!empty($servicestatus->get('perfdata'))) {
                    //Use parse perfdata string from database
                    $PerfdataParser = new PerfdataParser($servicestatus->get('perfdata'));
                    $perfdataMetadata = $PerfdataParser->parse();
                } else {
                    //Query graphite backend for available metrics - for example if service state is unknown
                    $metrics = $GraphiteLoader->findMetricsByUuid($hostUuid, $serviceUuid);

                    $perfdataMetadata = [];
                    foreach ($metrics as $metric) {
                        $perfdataMetadata[$metric] = [
                            'current'  => null,
                            'unit'     => null,
                            'warning'  => null,
                            'critical' => null,
                            'min'      => null,
                            'max'      => null
                        ];
                    }
                }

                $GraphiteLoader->setUseJsTimestamp($jsTimestamp);
                $GraphiteLoader->setAbsoluteDate($start, $end);

                foreach ($perfdataMetadata as $metricName => $metric) {
                    if ($gauge !== null) {
                        if ($gauge !== $metricName) {
                            continue;
                        }
                    }

                    $GraphiteMetric = new GraphiteMetric(
                        $hostUuid,
                        $serviceUuid,
                        $metricName
                    );

                    $datasource = [
                        'ds'    => $metricName,
                        'name'  => $metricName,
                        'label' => $metricName,
                        'unit'  => $metric['unit'],
                        'act'   => $metric['current'],
                        'warn'  => $metric['warning'],
                        'crit'  => $metric['critical'],
                        'min'   => $metric['min'],
                        'max'   => $metric['max'],
                    ];

                    if ($metric['unit'] === 'c') {
                        //Counter datatype
                        $performance_data[] = [
                            'datasource' => $datasource,
                            'data'       => $GraphiteLoader->getSeriesAsCounter($GraphiteMetric)
                        ];
                    } else {
                        //Default Gauge
                        switch ($type) {
                            case 'min':
                                $performance_data[] = [
                                    'datasource' => $datasource,
                                    'data'       => $GraphiteLoader->getSeriesMin($GraphiteMetric)
                                ];
                                break;

                            case 'max':
                                $performance_data[] = [
                                    'datasource' => $datasource,
                                    'data'       => $GraphiteLoader->getSeriesMax($GraphiteMetric)
                                ];
                                break;

                            default:
                                $performance_data[] = [
                                    'datasource' => $datasource,
                                    'data'       => $GraphiteLoader->getSeriesAvg($GraphiteMetric)
                                ];
                                break;
                        }
                    }
                }
            }
        }

        if ($this->PerfdataBackend->isMySQL()) {
            if (!empty($servicestatus)) {
                /** @var PerfdataTable $PerfdataTable */
                $PerfdataTable = TableRegistry::getTableLocator()->get('Perfdata');

                if (!empty($servicestatus->get('perfdata'))) {
                    //Use parse perfdata string from database
                    $PerfdataParser = new PerfdataParser($servicestatus->get('perfdata'));
                    $perfdataMetadata = $PerfdataParser->parse();
                } else {
                    //Query MySQL backend for available metrics - for example if service state is unknown - very slow!
                    $metrics = $PerfdataTable->findMetricsByUuid($hostUuid, $serviceUuid);

                    $perfdataMetadata = [];
                    foreach ($metrics as $metric) {
                        $metricName = $metric['label'];
                        $perfdataMetadata[$metricName] = [
                            'current'  => null,
                            'unit'     => null,
                            'warning'  => null,
                            'critical' => null,
                            'min'      => null,
                            'max'      => null
                        ];
                    }
                }

                foreach ($perfdataMetadata as $metricName => $metric) {
                    if ($gauge !== null) {
                        if ($gauge !== $metricName) {
                            continue;
                        }
                    }

                    $perfdataRaw = $PerfdataTable->getServicePerfdata(
                        $hostUuid,
                        $serviceUuid,
                        $metricName,
                        $start,
                        $end
                    );

                    $perfdata = [];
                    foreach ($perfdataRaw as $record) {
                        if ($debug === true) {
                            $perfdata[date('d.m.Y H:i:s', $record['timestamp'])] = $record['value'];
                        } else {
                            if ($jsTimestamp) {
                                $perfdata[$record['timestamp']] = $record['value'];
                            } else {
                                $perfdata[$record['timestamp_unix']] = $record['value'];
                            }
                        }
                    }

                    $datasource = [
                        'ds'    => $metricName,
                        'name'  => $metricName,
                        'label' => $metricName,
                        'unit'  => $metric['unit'],
                        'act'   => $metric['current'],
                        'warn'  => $metric['warning'],
                        'crit'  => $metric['critical'],
                        'min'   => $metric['min'],
                        'max'   => $metric['max'],
                    ];

                    //Default Gauge
                    switch ($type) {
                        case 'min':
                            $performance_data[] = [
                                'datasource' => $datasource,
                                'data'       => $this->getSeriesMin($perfdata)
                            ];
                            break;

                        case 'max':
                            $performance_data[] = [
                                'datasource' => $datasource,
                                'data'       => $this->getSeriesMax($perfdata)
                            ];
                            break;

                        default:
                            $performance_data[] = [
                                'datasource' => $datasource,
                                'data'       => $this->getSeriesAvg($perfdata)
                            ];
                            break;
                    }

                }
            }

        }

        if ($scale === true) {
            foreach ($performance_data as $index => $gauge) {
                $UnitScaler = new UnitScaler($gauge);

                if ($forcedUnit === null) {
                    $performance_data[$index] = $UnitScaler->scale();
                } else {
                    $performance_data[$index] = $UnitScaler->scaleToUnit($forcedUnit);
                }
            }
        }

        return $performance_data;
    }

    /**
     * @param array $perfdata
     * @return array
     */
    public function getSeriesAvg(array $perfdata) {
        if (sizeof($perfdata) < self::MAX_RESPONSE_GRAPH_POINTS) {
            return $perfdata;
        }

        $chunkSize = sizeof($perfdata) / self::MAX_RESPONSE_GRAPH_POINTS;
        $chunkSize = (int)$chunkSize;

        $result = [];
        foreach (array_chunk($perfdata, $chunkSize, true) as $chunk) {
            $currentChunkSize = sizeof($chunk);

            $average = array_sum($chunk) / $currentChunkSize;
            $averageTimestamp = array_sum(array_keys($chunk)) / $currentChunkSize;

            $result[$averageTimestamp] = $average;
        }

        return $result;
    }

    /**
     * @param array $perfdata
     * @return array
     */
    public function getSeriesMin(array $perfdata) {
        if (sizeof($perfdata) < self::MAX_RESPONSE_GRAPH_POINTS) {
            return $perfdata;
        }

        $chunkSize = sizeof($perfdata) / self::MAX_RESPONSE_GRAPH_POINTS;
        $chunkSize = (int)$chunkSize;

        $result = [];
        foreach (array_chunk($perfdata, $chunkSize, true) as $chunk) {
            $min = min($chunk);
            $timestamp = array_search($min, $chunk);

            $result[$timestamp] = $min;
        }

        return $result;
    }

    /**
     * @param array $perfdata
     * @return array
     */
    public function getSeriesMax(array $perfdata) {
        if (sizeof($perfdata) < self::MAX_RESPONSE_GRAPH_POINTS) {
            return $perfdata;
        }

        $chunkSize = sizeof($perfdata) / self::MAX_RESPONSE_GRAPH_POINTS;
        $chunkSize = (int)$chunkSize;

        $result = [];
        foreach (array_chunk($perfdata, $chunkSize, true) as $chunk) {
            $max = max($chunk);
            $timestamp = array_search($max, $chunk);

            $result[$timestamp] = $max;
        }

        return $result;
    }

}
