<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Console;


use itnovum\openITCOCKPIT\Checks\Receiver\CheckConfig;
use itnovum\openITCOCKPIT\Checks\Receiver\HttpLoader;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosHost;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\Plugins\Agent;
use itnovum\openITCOCKPIT\Checks\Receiver\Plugins\PluginInterface;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\TargetHost;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

class Poller extends Command {

    protected function configure() {
        $this
            ->setName('poller')
            ->setDescription('Poll JSON-data from openITCOCKPIT-Agent')
            ->setHelp("This command will poll JSON-data from openITCOCKPIT-Agent");

        $this->addOption(
            '--hostname',
            '-H',
            InputOption::VALUE_REQUIRED,
            'Host name or IP Address'
        );

        $this->addOption(
            '--http',
            null,
            InputOption::VALUE_NONE,
            'Disable warning for plaintext http connections. HTTPS connections are not affected by this option!'
        );

        $this->addOption(
            '--certificate-authority',
            '-C',
            InputOption::VALUE_OPTIONAL,
            'Path to CA file',
            '/opt/openitc/agent/server_ca.pem'
        );

        $this->addOption(
            '--certificate-authority-key',
            '-K',
            InputOption::VALUE_OPTIONAL,
            'Path to CA key file',
            '/opt/openitc/agent/server_ca.key'
        );

        $this->addOption(
            '--config',
            '-c',
            InputOption::VALUE_OPTIONAL,
            'Path to JSON configuration file with host and service check definitions',
            null
        );

        $this->addOption(
            '--german-host',
            null,
            InputOption::VALUE_OPTIONAL,
            'Address of the Gearman-Job-Server',
            '127.0.0.1'
        );

        $this->addOption(
            '--german-port',
            null,
            InputOption::VALUE_OPTIONAL,
            'Port of the Gearman-Job-Server',
            '4730'
        );

        $this->addOption(
            '--queryhandler',
            null,
            InputOption::VALUE_OPTIONAL,
            'Path to Naemon query handler',
            '/opt/openitc/nagios/var/rw/nagios.qh'
        );

        $this->addOption(
            '--dryrun',
            null,
            InputOption::VALUE_NONE,
            'Do not submit any results to the monitoring engine'
        );

        $this->addOption(
            '--no-proxy',
            null,
            InputOption::VALUE_NONE,
            'Do not use any HTTP/HTTPS proxy at all'
        );

        $this->addOption(
            '--keep-proxy-env',
            null,
            InputOption::VALUE_NONE,
            'Do not remove all proxy settings from environment (HTTP_PROXY, HTTPS_PROXY, etc)'
        );

        $this->addOption(
            '--host-check',
            null,
            InputOption::VALUE_NONE,
            'Determines of this command is used as an host check.'
        );

        $this->addOption(
            '--timeout',
            '-t',
            InputOption::VALUE_OPTIONAL,
            'Seconds before connection times out',
            30
        );

        $this->addOption(
            '--use-qh',
            null,
            InputOption::VALUE_NONE,
            'Use the Naemon Query Handler instead of Statusengine Broker to submit check results.'
        );

    }

    public function execute(InputInterface $input, OutputInterface $output) {
        $hostname = $input->getOption('hostname');
        $CACertPath = $input->getOption('certificate-authority');
        $CAKeyPath = $input->getOption('certificate-authority-key');
        $allowHttp = $input->getOption('http') === true;
        $queryHandler = $input->getOption('queryhandler');
        $useQueryHandler = $input->getOption('use-qh') === true;

        $gearmanHost = $input->getOption('german-host');
        $gearmanPort = $input->getOption('german-port');

        $dryRun = $input->getOption('dryrun') === true;
        $noProxy = $input->getOption('no-proxy') === true;
        $keepProxyEnv = $input->getOption('keep-proxy-env') === true;
        $isHostcheck = $input->getOption('host-check') === true;

        $timeout = (int)$input->getOption('timeout');
        if ($timeout <= 0) {
            $timeout = 30;
        }

        if ($keepProxyEnv === false) {
            $this->removeAllProxySettingsFromEnv();
        }

        if (!file_exists($queryHandler) && $dryRun === false) {
            if ($isHostcheck) {
                NagiosHost::doExit(
                    sprintf('Naemon Query Handler "%s" does not exists', $queryHandler),
                    NagiosHost::STATUS_UNREACHABLE,
                    true
                );
            } else {
                NagiosService::doExit(
                    sprintf('Naemon Query Handler "%s" does not exists', $queryHandler),
                    NagiosService::STATUS_UNKNOWN,
                    true
                );
            }
        }

        try {
            $configPath = $input->getOption('config');


            $CheckConfig = new CheckConfig($configPath);
            $config = $CheckConfig->getConfigByHostName($hostname);

            if ($config['mode'] !== 'pull') {
                if ($isHostcheck) {
                    NagiosHost::doExit(
                        'This host is configured to use PUSH Mode.',
                        NagiosHost::STATUS_DOWN
                    );
                } else {
                    NagiosService::doExit(
                        'This host is configured to use PUSH Mode.',
                        NagiosService::STATUS_CRITICAL
                    );
                }
            }

            $TargetHost = new TargetHost($config);
        } catch (\RuntimeException $e) {
            NagiosService::doExit(
                $e->getMessage(),
                NagiosService::STATUS_UNKNOWN,
                true
            );
        }

        $responseTime = 0.0;
        try {
            $proxy = $TargetHost->getProxyAddress();
            if ($noProxy === true) {
                //Proxy is disabled via CLI --no-proxy
                $proxy = false;
            }

            $HttpLoader = new HttpLoader($TargetHost, $CACertPath, $CAKeyPath, $proxy, $timeout);
            $start = microtime(true);
            $json = $HttpLoader->getResults();
            $responseTime = microtime(true) - $start;
        } catch (\Exception $e) {
            if ($isHostcheck) {
                NagiosHost::doExit(
                    $e->getMessage(),
                    NagiosHost::STATUS_UNREACHABLE,
                    true
                );
            } else {
                NagiosService::doExit(
                    $e->getMessage(),
                    NagiosService::STATUS_UNKNOWN,
                    true
                );
            }

        }

        if (!$TargetHost->hasChecks()) {
            if ($isHostcheck) {
                NagiosHost::doExit(
                    'No Agent checks defined for this host.',
                    NagiosHost::STATUS_DOWN,
                    true
                );
            } else {
                NagiosService::doExit(
                    'No Agent checks defined for this host.',
                    NagiosService::STATUS_UNKNOWN,
                    true
                );
            }
        }


        $checksConfig = $TargetHost->getChecksConfig();

        $results = [
            'Messages' => []
        ];
        if ($useQueryHandler === false) {
            $GearmanClient = new \GearmanClient();
            $GearmanClient->addServer($gearmanHost, $gearmanPort);
        }

        foreach ($checksConfig as $pluginConfig) {
            $pluginName = $pluginConfig['plugin'];

            $pluginClassName = sprintf('itnovum\openITCOCKPIT\Checks\Receiver\Plugins\%s', $pluginName);
            if (!class_exists($pluginClassName)) {
                if ($isHostcheck) {
                    NagiosHost::doExit(
                        sprintf('Unknown Plugin "%s"', $pluginName),
                        NagiosHost::STATUS_DOWN,
                        true
                    );
                } else {
                    NagiosService::doExit(
                        sprintf('Unknown Plugin "%s"', $pluginName),
                        NagiosService::STATUS_UNKNOWN,
                        true
                    );
                }
            }

            /** @var  $Plugin PluginInterface */
            $Plugin = new $pluginClassName($pluginConfig, $json);

            if ($output->isVerbose()) {
                $headline = sprintf("\n*************** Plugin [%s] ***************\n", $pluginClassName);
                echo $headline;
                $output->write('<info>Output: </info>');
                $output->writeln($Plugin->getOutput());
                $output->write('<info>Perfdata: </info>');
                $output->writeln($Plugin->getPerfdataSerialized());
                $output->write('<info>Long output: </info>');
                $output->writeln($Plugin->getLongOutput());
                $output->write('<info>Status code: </info>');
                $output->writeln($Plugin->getStatuscode());

                echo str_repeat('*', strlen($headline) - 2);
                echo "\n\n";
            }

            $pluginOutput = $Plugin->getOutput();
            if (strlen($Plugin->getPerfdataSerialized()) > 0) {
                $pluginOutput .= '|' . $Plugin->getPerfdataSerialized();
            }

            $command = "#command run ";

            $fullOutput = $pluginOutput;
            $brokerOutput = $pluginOutput;
            if (substr($pluginOutput, -1) !== "\n") {
                $fullOutput .= "\n";
            }

            $fullOutput .= $Plugin->getLongOutput();

            $checkData = [
                $config['uuid'], //host_name
                $pluginConfig['uuid'], //service_description
                $Plugin->getStatuscode(),
                $fullOutput
            ];

            $command .= '[' . time() . '] PROCESS_SERVICE_CHECK_RESULT;' . implode(';', $checkData);
            $command .= "\0";

            if ($dryRun === false && $useQueryHandler === true) {
                //Submit to Naemon via Query Handler
                $nagiosQH = fsockopen('unix://' . $queryHandler, null, $errno, $errstr, 5);
                fwrite($nagiosQH, $command);
                $response = trim(stream_get_contents($nagiosQH));
                fclose($nagiosQH);
            }

            // Create bulk message for Statusengine Broker
            $results['messages'][] = [
                'Command' => 'check_result',
                'Data'    => [
                    'host_name'           => $config['uuid'],
                    'service_description' => $pluginConfig['uuid'],
                    'output'              => $brokerOutput,
                    'long_output'         => ($Plugin->getLongOutput() === '' ? null : $Plugin->getLongOutput()),
                    'check_type'          => 1, //https://github.com/naemon/naemon-core/blob/cec6e10cbee9478de04b4cf5af29e83d47b5cfd9/src/naemon/common.h#L330-L334
                    'return_code'         => $Plugin->getStatuscode(),
                    'start_time'          => time(),
                    'end_time'            => time(),
                    'early_timeout'       => 0,
                    'latency'             => 0,
                    'exited_ok'           => 1
                ]
            ];
        }

        if ($dryRun === false && $useQueryHandler === false) {
            //Submit to Naemon via the Statusengine Broker
            $GearmanClient->doBackground('statusngin_cmd', json_encode($results));
        }

        //This is the active check (executed my Naemon/Nagios), return agent data to Nagios/Naemon
        $Agent = new Agent([], $json, $responseTime);

        if ($TargetHost->useHttp() && $allowHttp === false) {
            if ($isHostcheck) {
                NagiosHost::doExit(
                    sprintf('Unencrypted HTTP connection used! You can disable this warning by passing --http. %s', $Agent->getOutput(false) . '|' . $Agent->getPerfdataSerialized()),
                    NagiosHost::STATUS_DOWN,
                    true
                );
            } else {
                NagiosService::doExit(
                    sprintf('Unencrypted HTTP connection used! You can disable this warning by passing --http. %s', $Agent->getOutput(false) . '|' . $Agent->getPerfdataSerialized()),
                    NagiosService::STATUS_WARNING,
                    true
                );
            }
        }

        if ($isHostcheck) {
            $rc = $Agent->getStatuscode();
            if ($rc > 0) {
                $rc = 1;
            }

            NagiosHost::doExit(
                $Agent->getOutput() . '|' . $Agent->getPerfdataSerialized(),
                $rc,
                false
            );
        }
        NagiosService::doExit(
            $Agent->getOutput() . '|' . $Agent->getPerfdataSerialized(),
            $Agent->getStatuscode(),
            false
        );
    }

    public function removeAllProxySettingsFromEnv() {
        $vars = [
            'HTTP_PROXY',
            'http_proxy',
            'HTTPS_PROXY',
            'https_proxy',
            'FTP_PROXY',
            'ftp_proxy',
            'RSYNC_PROXY',
            'rsync_proxy',
            'NO_PROXY',
            'no_proxy'
        ];
        foreach ($vars as $var) {
            putenv($var);
        }
    }

}
