<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Console;


use itnovum\openITCOCKPIT\Checks\Receiver\HttpLoader;
use itnovum\openITCOCKPIT\Checks\Receiver\Security\AgentCertificateData;
use itnovum\openITCOCKPIT\Checks\Receiver\Security\AgentHttpClient;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\TargetHost;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;

class Query extends Command {

    protected function configure() {
        $this
            ->setName('query')
            ->setDescription('Query JSON-data from openITCOCKPIT-Agent')
            ->setHelp("This command will query JSON-data from openITCOCKPIT-Agent");

        $this->addOption(
            '--host',
            '-H',
            InputOption::VALUE_REQUIRED,
            'FQDN or IP Address'
        );

        $this->addOption(
            '--port',
            '-P',
            InputOption::VALUE_OPTIONAL,
            'Port of the openITCOCKPIT Monitoring Agent',
            '3333'
        );

        $this->addOption(
            '--certificate-authority',
            '-C',
            InputOption::VALUE_OPTIONAL,
            'Path to CA file',
            '/opt/openitc/agent/server_ca.pem'
        );

        $this->addOption(
            '--certificate-authority-key',
            '-K',
            InputOption::VALUE_OPTIONAL,
            'Path to CA key file',
            '/opt/openitc/agent/server_ca.key'
        );

        $this->addOption(
            '--auto-tls',
            null,
            InputOption::VALUE_NONE,
            'openITCOCKPIT automatically generate a TLS certificate for authentication and encryption purpose.'
        );

        $this->addOption(
            '--https',
            null,
            InputOption::VALUE_NONE,
            'Query the Agent via HTTPS - using own certificates'
        );

        $this->addOption(
            '--insecure',
            '-k',
            InputOption::VALUE_NONE,
            'Disable TLS certificate validation. This requires valid certificates like from Let\'s Encrypt (Only with --https option)'
        );

        $this->addOption(
            '--http',
            null,
            InputOption::VALUE_NONE,
            'Communication will be plaintext. No encryption and authentication. Not recommended!'
        );

        $this->addOption(
            '--no-proxy',
            null,
            InputOption::VALUE_NONE,
            'Do not use any HTTP/HTTPS proxy at all'
        );

        $this->addOption(
            '--timeout',
            '-t',
            InputOption::VALUE_OPTIONAL,
            'Seconds before connection times out',
            30
        );

        $this->addOption(
            '--username',
            '-u',
            InputOption::VALUE_OPTIONAL,
            'Username for HTTP Basic Authentication',
            ''
        );

        $this->addOption(
            '--password',
            '-p',
            InputOption::VALUE_OPTIONAL,
            'Password for HTTP Basic Authentication',
            ''
        );

        $this->addOption(
            '--pretty',
            null,
            InputOption::VALUE_NONE,
            'Print pretty formatted JSON'
        );

        $this->addOption(
            '--test-connection-and-autotls',
            null,
            InputOption::VALUE_NONE,
            'Only test the connection to the given Agent and exchange Auto-TLS certificates if needed.'
        );

        $this->addOption(
            '--autossl-successful',
            null,
            InputOption::VALUE_NONE,
            'Set this if there was a successful exchange of Auto-TLS certificates in the past.'
        );

    }

    public function execute(InputInterface $input, OutputInterface $output) {
        $host = $input->getOption('host');
        $port = (int)$input->getOption('port');
        if ($port <= 0 || $port > 65535) {
            $port = 3333;
        }

        $CACertPath = $input->getOption('certificate-authority');
        $CAKeyPath = $input->getOption('certificate-authority-key');
        $autoTls = $input->getOption('auto-tls') === true;

        $https = $input->getOption('https') === true;
        $insecure = $input->getOption('insecure') === true;

        $http = $input->getOption('http') === true;

        $useBasicAuth = false;
        $username = $input->getOption('username');
        $password = $input->getOption('password');
        if (strlen($username) > 0 && strlen($password) > 0) {
            $useBasicAuth = true;
        }

        $pretty = $input->getOption('pretty') === true;

        $onlyTestConnectionAndExchangeAutoTls = $input->getOption('test-connection-and-autotls') === true;
        $autossl_successful = $input->getOption('autossl-successful') === true;

        $useHttps = $https;
        if ($http === true) {
            // User passed --http option
            $useHttps = false;
        }

        $noProxy = $input->getOption('no-proxy') === true;

        //Use Auto-TLS as default
        if (!$autoTls && !$https && !$http) {
            $autoTls = true;
        }

        $timeout = (int)$input->getOption('timeout');
        if ($timeout <= 0) {
            $timeout = 30;
        }

        $TargetHost = new TargetHost([
            'address'     => $host,
            'port'        => $port,
            'uuid'        => 'FAKEUUID-HARD-CODE-NOTT-REQUIREDDDDD', //Not required for querying the agent
            'name'        => 'HARD-CODE-HOST-NAME-NOT-REQUIRED', //Not required for querying the agent
            'use_autossl' => $autoTls,
            'use_https'   => $useHttps,
            'insecure'    => $insecure,
            'basic_auth'  => $useBasicAuth,
            'username'    => $username,
            'password'    => $password,
            'mode'        => 'pull', // If we query the agent manually via this script - it is definitely pull :)
            'proxy'       => $this->getProxyFromEnv(),
            'checks'      => [], // Not required
        ]);

        /** @var AgentCertificateData $AgentCertificateData */
        $AgentCertificateData = new AgentCertificateData();
        if ($AgentCertificateData->serverCAExists() === false) {
            $AgentCertificateData->generateServerCA();
        }

        $AgentHttpClient = new AgentHttpClient($TargetHost);
        $result = $AgentHttpClient->testConnectionAndExchangeAutotls($autossl_successful);

        if ($result['status'] !== 'success') {
            if ($pretty) {
                echo json_encode($result, JSON_PRETTY_PRINT);
            } else {
                //Print out agent result
                echo json_encode($result);
            }
            exit(1);
        } else {
            if ($onlyTestConnectionAndExchangeAutoTls) {
                if ($pretty) {
                    echo json_encode($result, JSON_PRETTY_PRINT);
                } else {
                    //Print out agent result
                    echo json_encode($result);
                }
                exit(0);
            }
        }

        try {
            $proxy = $TargetHost->getProxyAddress();
            if ($noProxy === true) {
                //Proxy is disabled via CLI --no-proxy
                $proxy = false;
            }

            $HttpLoader = new HttpLoader($TargetHost, $CACertPath, $CAKeyPath, $proxy, $timeout);
            $json = $HttpLoader->getResults();

            if ($pretty) {
                echo json_encode($json, JSON_PRETTY_PRINT);
            } else {
                //Print out agent result
                echo json_encode($json);
            }
        } catch (\Exception $e) {
            echo json_encode([
                'status'  => 'error',
                'message' => $e->getMessage()
            ], JSON_PRETTY_PRINT);
            exit(1);
        }

    }

    /**
     * @return false|string
     */
    public function getProxyFromEnv() {
        $vars = [
            'HTTP_PROXY',
            'http_proxy',
            'HTTPS_PROXY',
            'https_proxy'
        ];
        foreach ($vars as $var) {
            $proxy = getenv($var);

            if ($proxy) {
                return $proxy;
            }
        }

        return false;
    }

}
