<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver;


use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;
use itnovum\openITCOCKPIT\Checks\Exceptions\LoadingException;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\TargetHost;

class HttpLoader {

    /**
     * @var TargetHost
     */
    private $TargetHost;

    /**
     * @var string
     */
    private $caCertFile;

    /**
     * @var string
     */
    private $caKeyFile;

    /**
     * @var false|string
     */
    private $proxy;

    /**
     * @var int
     */
    private $timeout = 30;

    /**
     * HttpLoader constructor.
     * @param TargetHost $TargetHost
     * @param string $caCertFile
     * @param string $caKeyFile
     * @param bool $proxy
     * @param int $timeout
     */
    public function __construct(TargetHost $TargetHost, string $caCertFile, string $caKeyFile, $proxy = false, $timeout = 30) {
        $this->TargetHost = $TargetHost;
        $this->caCertFile = $caCertFile;
        $this->caKeyFile = $caKeyFile;
        $this->proxy = $proxy;
        $this->timeout = $timeout;
    }

    /**
     * @return array
     * @throws LoadingException
     * @throws GuzzleException
     */
    public function getResults() {
        $options = $this->getGuzzleOptions();
        $url = sprintf('%s/', $this->getBaseUrl());
        $client = new Client();

        $response = $client->request('GET', $url, $options);
        if ($response->getStatusCode() === 200) {
            $data = @json_decode($response->getBody()->getContents(), true);
            if (json_last_error() === JSON_ERROR_NONE && isset($data['agent'])) {
                return $data;
            }

            throw new LoadingException('Agent response is not a json or has no agent key');
        }

        throw new LoadingException(sprintf('[%s] %s', $response->getStatusCode(), $response->getBody()->getContents()));
    }

    /**
     * Get the Guzzle connection options for ALL requests to the openITCOCKPIT Monitoring Agent
     * @return array
     */
    public function getGuzzleOptions() {
        $options = [
            'headers' => [
                'Content-Type' => 'application/json'
            ],
            'verify'  => true,
            'proxy'   => [
                'http'  => false,
                'https' => false
            ],
            'timeout' => $this->timeout
        ];

        // Add Proxy options
        if ($this->proxy) {
            $options['proxy'] = [
                'http'  => $this->proxy,
                'https' => $this->proxy
            ];
        }

        // HTTP Basic Auth
        if ($this->TargetHost->useBasicAuth() === true) {
            $options['auth'] = [
                $this->TargetHost->getUsername(),
                $this->TargetHost->getPassword()
            ];
        }

        // Disable HTTPS cert validation?
        if ($this->TargetHost->useHttps() === true && $this->TargetHost->allowInsecureHttps() === true) {
            $options['verify'] = false;
        }

        // Use AutoTLS?
        if ($this->TargetHost->useAutoTLS() === true) {
            // Disable hostname validation
            // cURL error 51: SSL: certificate subject name 'subject' does not match target host name '127.0.0.1' (see https://curl.haxx.se/libcurl/c/libcurl-errors.html)
            //$options['verify'] = $AgentCertificateData->getCaCertFile();
            $options['verify'] = false;

            $options['cert'] = $this->caCertFile;
            $options['ssl_key'] = $this->caKeyFile;
        }

        return $options;
    }

    /**
     * @return string
     */
    private function getBaseUrl() {
        $protocol = 'https';
        // Is HTTP (plaintext) enabled?
        if ($this->TargetHost->useHttp()) {
            $protocol = 'http';
        }

        // e.g.: https://127.0.0.1:3333
        return sprintf(
            '%s://%s:%s',
            $protocol,
            $this->TargetHost->getAddress(),
            $this->TargetHost->getPort()
        );
    }

}
