<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

abstract class BasePlugin implements PluginInterface {

    /**
     * @var array
     */
    protected $pluginConfig = [];

    /**
     * @var array
     */
    protected $json;

    /**
     * @var string
     */
    protected $output = '';

    /**
     * @var string
     */
    protected $longOutput = '';

    /**
     * @var int
     */
    protected $statusCode = 3;

    /**
     * @var array
     */
    private $perfdata = [];

    /**
     * BasePlugin constructor.
     * @param array $pluginConfig
     * @param array $json
     */
    public function __construct($pluginConfig, $json) {
        $this->pluginConfig = $pluginConfig;
        $this->json = $json;
        $this->execute();
    }

    /**
     * @param $statusCode
     * @return $this
     */
    protected function setStatuscode($statusCode) {
        $this->statusCode = (int)$statusCode;
        return $this;
    }

    /**
     * @param $output
     * @return $this
     */
    protected function setOutput($output) {
        $this->output = $output;
        return $this;
    }

    /**
     * @param string $longOutput
     * @return $this
     */
    protected function setLongOutput($longOutput) {
        $this->longOutput = $longOutput;
        return $this;
    }

    /**
     * @param string $longOutput
     * @param bool $addNewLine
     * @return $this
     */
    protected function appendLongOutput($longOutput, $addNewLine = true) {
        $this->longOutput .= $longOutput;
        if ($addNewLine) {
            $this->longOutput .= PHP_EOL;
        }
        return $this;
    }

    /**
     * @param PerfdataGauge $PerfdataGauge
     * @return $this
     */
    protected function addPerfdata(PerfdataGauge $PerfdataGauge) {
        $this->perfdata[] = $PerfdataGauge;
        return $this;
    }

    /**
     * @return int
     */
    public function getStatuscode() {
        return $this->statusCode;
    }

    /**
     * @param bool $doStatusPrefix
     * @return string
     */
    public function getOutput($doStatusPrefix = true) {
        if ($doStatusPrefix) {
            return NagiosService::prefixOutput($this->output, $this->statusCode);
        }

        return $this->output;
    }

    /**
     * @return string
     */
    public function getLongOutput() {
        return $this->longOutput;
    }

    /**
     * @return string
     */
    public function getPerfdataSerialized() {
        if (empty($this->perfdata)) {
            return '';
        }

        $perfdataStrings = [];

        foreach ($this->perfdata as $Gauge) {
            /** @var PerfdataGauge $Gauge */
            $perfdataString = sprintf(
                '%s=%s%s;%s;%s;%s;%s',
                $Gauge->getLabel(),
                $Gauge->getValue(),
                $Gauge->getUnit(),
                $Gauge->getWarning(),
                $Gauge->getCritical(),
                $Gauge->getMin(),
                $Gauge->getMax()
            );
            $perfdataStrings[] = $perfdataString;
        }

        return implode(' ', $perfdataStrings);
    }

    /**
     * Returns false, even if $threshold is an empty string.
     * If $value fits the defined range in $threshold, returns true.
     *
     * @param $threshold
     * @param $value
     * @return bool
     */
    public function getRangeStatus($threshold, $value) {
        $value = floatval($value);

        // Ranges: https://www.monitoring-plugins.org/doc/guidelines.html#THRESHOLDFORMAT
        // | Range definition |            Generate an alert if x...             |  Case  |
        // |:----------------:|:------------------------------------------------:|:-------|
        // | 10               | < 0 or > 10, (outside the range of {0 .. 10})    |   1    |
        // | 10:              | < 10, (outside {10 ... ∞})                       |   2    |
        // | ~:10             | > 10, (outside the range of {-∞ ... 10})         |   3    |
        // | 10:20            | < 10 or > 20, (outside the range of {10 ... 20}) |   4    |
        // | @10:20           | ≥ 10 and ≤ 20, (inside the range of {10 ... 20}) |   5    |


        $threshold = trim($threshold);
        if (empty($threshold)) {
            return false;
        }

        // Is this a range at all? (case 1)
        if (strpos($threshold, ':') === false) {
            $max = floatval($threshold);
            return ($value < 0 || $value > $max);
        }

        $ranges = explode(':', $threshold, 2);
        $min = $ranges[0];
        $max = $ranges[1] ?? null;

        // case 3
        if ($min === '~') {
            return $value > floatval($max);
        }

        // case 2
        if (empty($max)) {
            return $value < floatval($min);
        }

        // If the $threshold starts with an "@" it is a inside of the range threshold
        $insideRange = strpos($min, '@') === 0;
        // also remote the "@"
        $min = substr($min, 1);

        if ($insideRange === true) {
            // case 5
            return $value >= floatval($min) && $value <= floatval($max);
        }

        // case 4
        return $value < floatval($min) || $value > floatval($max);
    }

    /**
     * @param string $key in agent json
     * @return bool true if an error occurred
     */
    public function checkError($mainKey, $subKey = null) {
        if (!empty($subKey)) {
            if (isset($this->json[$mainKey][$subKey]['error'])) {
                $this->setOutput(sprintf(
                    'Agent error: %s',
                    $this->json[$mainKey][$subKey]['error']
                ));
                $this->setStatuscode(NagiosService::STATUS_CRITICAL);
                return true;
            }
        }

        if (isset($this->json[$mainKey]['error'])) {
            $this->setOutput(sprintf(
                'Agent error: %s',
                $this->json[$mainKey]['error']
            ));
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return true;
        }
        return false;
    }

}
