<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\Converter\Byte;
use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;
use itnovum\openITCOCKPIT\Checks\Receiver\ValueObjects\PerfdataGauge;

/**
 * Class Libvirt
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class Libvirt extends BasePlugin implements PluginInterface {

    public function execute() {
        if ($this->checkError('net_io')) {
            return;
        }

        $requiredKeys = [
            'libvirt' => [
                'Name',
                'Uuid',
                'Memory',
                'CpuUsage'
            ]
        ];

        foreach ($requiredKeys as $mainKey => $subKeys) {
            if (isset($this->json[$mainKey])) {
                foreach ($this->json[$mainKey] as $vm => $data) {
                    foreach ($subKeys as $subKey) {
                        if (!array_key_exists($subKey, $this->json[$mainKey][$vm])) { // values can be null which is not supported by isset()
                            $this->setOutput(sprintf('Key "%s.%s.%s" not found in Agent output', $mainKey, $vm, $subKey));
                            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
                            return;
                        }
                    }
                }
            }
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $uuid = '';
        // Percentage values
        $warningMemory = 80;
        $criticalMemory = 90;
        $warningCpu = 80;
        $criticalCpu = 90;
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $uuid = $arg;
                    break;
                case 1:
                    $warningMemory = (int)$arg;
                    break;
                case 2:
                    $criticalMemory = (int)$arg;
                    break;
                case 3:
                    $warningCpu = (int)$arg;
                    break;
                case 4:
                    $criticalCpu = (int)$arg;
                    break;
            }
        }

        if (!isset($this->json['libvirt'][$uuid])) {
            $this->setOutput(sprintf('VM "%s" not found in Agent output', $uuid));
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $vm = $this->json['libvirt'][$uuid];

        if (!$vm['IsRunning']) {
            // VM is not running
            $this->setOutput(sprintf('VM %s is not running', $vm['Name']));
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);

            $this->addPerfdata(new PerfdataGauge(
                'cpu_usage',
                0,
                '%',
                $warningCpu,
                $criticalCpu,
                0,
                100
            ));

            $this->addPerfdata(new PerfdataGauge(
                'memory_usage_bytes',
                0,
                'B'
            ));

            $this->addPerfdata(new PerfdataGauge(
                'memory_total_bytes',
                0,
                'B'
            ));

            $this->addPerfdata(new PerfdataGauge(
                'memory_usage',
                0,
                '%',
                $warningMemory,
                $criticalMemory,
                0,
                100
            ));

            return;
        }

        //All memory values are in kb
        $TotalBytes = $vm['Memory']['Total'] * 1024;
        $RssBytes = $vm['Memory']['Rss'] * 1024;

        $usedMemoryPercentage = $RssBytes / $TotalBytes * 100;

        $this->setStatuscode(NagiosService::STATUS_OK);
        $this->setOutput(sprintf(
            'VM %s is is running. CPU Usage: %s%% Memory usage: %s/%s (%s%%)',
            $vm['Name'],
            round($vm['CpuUsage']['GuestPercent'], 2),
            Byte::asHuman($RssBytes),
            Byte::asHuman($TotalBytes),
            round($usedMemoryPercentage, 2)
        ));

        $this->addPerfdata(new PerfdataGauge(
            'cpu_usage',
            round($vm['CpuUsage']['GuestPercent'], 2),
            '%',
            $warningCpu,
            $criticalCpu,
            0,
            100
        ));

        $this->addPerfdata(new PerfdataGauge(
            'memory_usage_bytes',
            $RssBytes,
            'B'
        ));

        $this->addPerfdata(new PerfdataGauge(
            'memory_total_bytes',
            $TotalBytes,
            'B'
        ));

        $this->addPerfdata(new PerfdataGauge(
            'memory_usage',
            round($usedMemoryPercentage, 2),
            '%',
            $warningMemory,
            $criticalMemory,
            0,
            100
        ));

        if ($vm['CpuUsage']['GuestPercent'] >= $criticalCpu) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($vm['CpuUsage']['GuestPercent'] >= $warningCpu) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }

        if ($usedMemoryPercentage >= $criticalMemory) {
            $this->setStatuscode(NagiosService::STATUS_CRITICAL);
            return;
        }

        if ($usedMemoryPercentage >= $warningMemory) {
            $this->setStatuscode(NagiosService::STATUS_WARNING);
            return;
        }
    }
}
