<?php
// Copyright (C) <2018>  <it-novum GmbH>
//
// This file is dual licensed
//
// 1.
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, version 3 of the License.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// 2.
//  If you purchased an openITCOCKPIT Enterprise Edition you can use this file
//  under the terms of the openITCOCKPIT Enterprise Edition license agreement.
//  License agreement and license key will be shipped with the order
//  confirmation.

namespace itnovum\openITCOCKPIT\Checks\Receiver\Plugins;


use itnovum\openITCOCKPIT\Checks\Receiver\NagiosService;

/**
 * Class SystemdService
 * @package itnovum\openITCOCKPIT\Checks\Receiver\Plugins
 */
class SystemdService extends BasePlugin implements PluginInterface {


    public function execute() {
        if ($this->checkError('systemd_services')) {
            return;
        }

        if (!isset($this->json['systemd_services']) || !is_array($this->json['systemd_services'])) {
            $this->setOutput('Key "systemd_services" not found in Agent output');
            return;
        }

        if (!isset($this->pluginConfig['args'])) {
            $this->setOutput('Plugin arguments not found in configuration');
            $this->setStatuscode(NagiosService::STATUS_UNKNOWN);
            return;
        }

        $unit = null;
        $strict = true;
        foreach ($this->pluginConfig['args'] as $key => $arg) {
            switch ($key) {
                case 0:
                    $unit = trim($arg);
                    break;
                case 1:
                    $strict = (bool)$arg;
                    break;
            }
        }

        foreach ($this->json['systemd_services'] as $service) {
            if ($this->serviceUnitMatch($unit, $service['Name'], $strict)) {
                $this->setOutput(sprintf('%s %s (%s)', $service['Name'], $service['LoadState'], $service['SubState']));
                $longOutput = [
                    sprintf('Name: %s', $service['Name']),
                    sprintf('Description: %s', $service['Description']),
                    sprintf('Loaded: %s', $service['LoadState']),
                    sprintf('Active: %s', $service['ActiveState']),
                    sprintf('State: %s', $service['SubState']),
                ];

                $this->setLongOutput(implode(PHP_EOL, $longOutput));

                /*
                 * States for .service units
                 * Source: https://www.computernetworkingnotes.com/linux-tutorials/systemd-units-explained-with-types-and-states.html
                 *
                 * loaded             The unit's configuration file has been successfully read and processed.
                 * Active (exited)    Successfully executed the one-time configuration and after execution, the unit is neither running an active process nor waiting for an event.
                 * Active (running)   Successfully executed the one-time configuration and after execution, the unit is running one or more active processes.
                 * Active (waiting)   Successfully executed the one-time configuration and after execution, the unit is waiting for an event.
                 * Inactive (dead)    Either the one-time configuration failed to execute or not executed yet.
                 *
                 * ----
                 * A .target unit like "nfs-client.target" can have an "active" state
                 */

                if ($service['LoadState'] == 'loaded' && $service['SubState'] != 'dead') {
                    $this->setStatuscode(NagiosService::STATUS_OK);

                    if ($service['SubState'] != 'exited' && $service['SubState'] != 'running' && $service['SubState'] != 'waiting' && $service['SubState'] != 'active') {
                        // Not sure if we will ever enter this if
                        $this->setStatuscode(NagiosService::STATUS_WARNING);
                    }
                } else {
                    $this->setStatuscode(NagiosService::STATUS_CRITICAL);
                }
                return;
            }
        }

        $this->setOutput(sprintf('Service (%s) not found', $unit));
        $this->setStatuscode(NagiosService::STATUS_CRITICAL);
    }

    private function serviceUnitMatch($match, $unit, $strict) {
        if ($strict && $match === $unit) {
            return true;
        }
        if (!$strict && strpos($unit, $match) !== false) {
            return true;
        }
        return false;
    }
}
